import os,sys,glob,time,shutil,argparse,json,traceback,copy,re
from . import globV,comm
from dflow import (
    Workflow,
    Step,
    Steps,
    Inputs,
    Outputs,
    argo_range,
    SlurmRemoteExecutor,
    upload_artifact,
    download_artifact,
    InputArtifact,
    InputParameter,
    OutputArtifact,
    OutputParameter,
    ShellOPTemplate,
    S3Artifact,
    argo_len,
    argo_sequence,
)

from pathlib import Path
from typing import List, Optional

from dflow.python import (
    PythonOPTemplate,
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    Slices,
    BigParameter,
    Parameter
)

from . import comm,metrics,tracking

def execute_postdft(work_path,allexample_path,extra_files,extra_files_art_root,tracking_setting,metrics_setting,super_metrics_setting,command,outputfiles):
    """
    should execute this function in abacustest submited path.
    
    
    work_path: str, work path, should be a absolute path
    allexample_path: list, all example path, should be relative path to work_path
    extra_files: should be a list of extra files, and relative path to current path
    
    
    """
    
    #copy extra_files to work path 
    if extra_files:
        for iextra in extra_files:
            comm.CopyFiles(iextra,work_path,move=False)
    if extra_files != None:
        extra_file_path = extra_files_art_root
        comm.CopyFiles(extra_file_path,work_path,move=False)
    #check if need to upload to tracking
    do_upload_tracking = False
    if tracking_setting and tracking_setting.get("ifurn",True):
        do_upload_tracking = True
        
    #read metrics
    metrics_setting_list = metrics.Metrics.TransferMetricsOPIO(metrics_setting)
    # need to split metrics_setting_list to before_command and after_command
    metrics_setting_list_before_command = []
    metrics_setting_list_after_command = []
    for imetric in metrics_setting_list:
        if imetric.get("before_command",True):
            metrics_setting_list_before_command.append(imetric)
        else:
            metrics_setting_list_after_command.append(imetric)
    tracking_values_before = tracking_values_after = None
    if metrics_setting_list_before_command:
        try:
            os.chdir(work_path)
            tracking_values_before = metrics.ReadMetrics(metrics_setting_list_before_command,do_upload_tracking,allexample_path)
        except:
            traceback.print_exc()
    
    #execute command
    os.chdir(work_path)
    log = f"COMMAND: {command}"
    if command.strip() != "":
        cmd = str(command)
        return_code, out, err = comm.run_command(cmd)
        log += out + err
        #log += os.popen("(%s) 2>&1" % cmd).read()
    
    if metrics_setting_list_after_command:
        try:
            os.chdir(work_path)
            tracking_values_after = metrics.ReadMetrics(metrics_setting_list_after_command,do_upload_tracking,allexample_path)
        except:
            traceback.print_exc()              
    # merge tracking_values_before and tracking_values_after
    tracking_values = []
    if tracking_values_before:
        tracking_values.extend(tracking_values_before)
    if tracking_values_after:
        tracking_values.extend(tracking_values_after)
    if not tracking_values:
        tracking_values = None
    #calculate super_metrics
    try:
        os.chdir(work_path)
        tracking_summary,report = metrics.ReadSuperMetrics(metrics.Metrics.TransferMetricsOPIO(super_metrics_setting),do_upload_tracking)  
    except:
        traceback.print_exc()
        tracking_summary = None  
                
    #upload tracking
    if do_upload_tracking:
        try:
            tracking.upload_to_tracking(tracking_setting,tracking_values,tracking_summary,AIM_ACCESS_TOKEN=None)
        except:
            traceback.print_exc() 
              
    # get cpuinfo to CPUINFO.log
    os.chdir(work_path)
    try:
        os.system("lscpu > CPUINFO.log")
    except:
        pass
    cpuinfo_log = None if not os.path.isfile("CPUINFO.log") else "CPUINFO.log" 
    if cpuinfo_log:
        with open(cpuinfo_log) as f1:
            log += "\n\nCPUINFO:\n" + f1.read() 
                            
    #collect outputs
    outpath = []
    os.chdir(work_path)
    logfile_name = "STDOUTER.log"
    i = 1
    while os.path.isfile(logfile_name):
        logfile_name = "STDOUTER_%d.log" % i
        i += 1
    logfile = Path(logfile_name)
    if len(outputfiles) == 0:
        for ifile in glob.glob("*"):
            if ifile.startswith("."):
                continue 
            outpath.append(Path(ifile))
    else:
        for i in outputfiles:
            for j in glob.glob(i):
                if os.path.exists(j):
                    outpath.append(Path(j))
                else:
                    log += "\n%s is not exist" % j
        outpath.append(logfile)
        if cpuinfo_log:
            outpath.append(Path(cpuinfo_log))
        if os.path.isfile("version.dat"):
            outpath.append(Path("version.dat"))
    #print("log:",log,file=sys.stderr)
    logfile.write_text(log)
    print("outpath:",str(outpath),file=sys.stderr)
    return outpath


class PostDFT(OP):
    def __init__(self):
        pass

    @classmethod
    def get_input_sign(cls):
        return OPIOSign(
            {
                "examples": Artifact(List[Path]),
                "command": str,
                "extra_files": Artifact(Path,optional=True),
                "outputfiles":[str],
                "metrics": BigParameter(dict,default={}),
                "super_metrics": BigParameter(dict,default={}),
                "upload_tracking": BigParameter(dict,default={}),
            }
        )

    @classmethod
    def get_output_sign(cls):
        return OPIOSign(
            {
                "outputs": Artifact(List[Path])
            }
        )

    @OP.exec_sign_check
    def execute(
        self,
        op_in: OPIO,
    ) -> OPIO:

        print("op_in:",op_in,file=sys.stderr)
        outpath = []
        
        # if define sub_save_path, create sub_save_path in root and copy examples to sub_save_path
        #example_path = str(op_in["examples"][0]).split("/inputs/artifacts/examples/")[0] + "/inputs/artifacts/examples/"
        #work_path = example_path
        work_path = op_in["examples"].art_root
        
        allexample_path = []
        for i in op_in["examples"]:
            #allexample_path.append(str(i).split("/inputs/artifacts/examples/")[1])
            allexample_path.append(os.path.relpath(str(i),str(work_path)))

        print("work path:",work_path,file=sys.stderr)
        
        #copy extra_files to work path 
        if op_in["extra_files"] != None:
            #extra_file_path = str(op_in["extra_files"]).split("/inputs/artifacts/extra_files")[0] + "/inputs/artifacts/extra_files"
            extra_file_path = op_in["extra_files"].art_root
            comm.CopyFiles(extra_file_path,work_path,move=False)

        #check if need to upload to tracking
        tracking_setting = op_in["upload_tracking"]
        metrics_setting = op_in["metrics"]
        super_metrics_setting = op_in["super_metrics"]
        do_upload_tracking = False
        if tracking_setting and tracking_setting.get("ifurn",True):
            do_upload_tracking = True
            
        #read metrics
        metrics_setting_list = metrics.Metrics.TransferMetricsOPIO(metrics_setting)
        # need to split metrics_setting_list to before_command and after_command
        metrics_setting_list_before_command = []
        metrics_setting_list_after_command = []
        for imetric in metrics_setting_list:
            if imetric.get("before_command",True):
                metrics_setting_list_before_command.append(imetric)
            else:
                metrics_setting_list_after_command.append(imetric)
        tracking_values_before = tracking_values_after = None
        if metrics_setting_list_before_command:
            try:
                os.chdir(work_path)
                tracking_values_before = metrics.ReadMetrics(metrics_setting_list_before_command,do_upload_tracking,allexample_path)
            except:
                traceback.print_exc()
        
        #execute command
        os.chdir(work_path)
        log = f"COMMAND: {op_in['command']}"
        if op_in["command"].strip() != "":
            cmd = "ulimit -c 0; " + str(op_in["command"])
            return_code, out, err = comm.run_command(cmd)
            log += out + err
            #log += os.popen("(%s) 2>&1" % cmd).read()
        
        if metrics_setting_list_after_command:
            try:
                os.chdir(work_path)
                tracking_values_after = metrics.ReadMetrics(metrics_setting_list_after_command,do_upload_tracking,allexample_path)
            except:
                traceback.print_exc()              

        # merge tracking_values_before and tracking_values_after
        tracking_values = []
        if tracking_values_before:
            tracking_values.extend(tracking_values_before)
        if tracking_values_after:
            tracking_values.extend(tracking_values_after)
        if not tracking_values:
            tracking_values = None
        #calculate super_metrics
        try:
            os.chdir(work_path)
            tracking_summary,report = metrics.ReadSuperMetrics(metrics.Metrics.TransferMetricsOPIO(super_metrics_setting),do_upload_tracking)  
        except:
            traceback.print_exc()
            tracking_summary = None  
                    
        #upload tracking
        if do_upload_tracking:
            try:
                tracking.upload_to_tracking(tracking_setting,tracking_values,tracking_summary,AIM_ACCESS_TOKEN=None)
            except:
                traceback.print_exc() 
                  
        # get cpuinfo to CPUINFO.log
        os.chdir(work_path)
        try:
            os.system("lscpu > CPUINFO.log")
        except:
            pass
        cpuinfo_log = None if not os.path.isfile("CPUINFO.log") else "CPUINFO.log" 
        if cpuinfo_log:
            with open(cpuinfo_log) as f1:
                log += "\n\nCPUINFO:\n" + f1.read() 
                                
        #collect outputs
        os.chdir(work_path)
        logfile_name = "STDOUTER.log"
        i = 1
        while os.path.isfile(logfile_name):
            logfile_name = "STDOUTER_%d.log" % i
            i += 1
        logfile = Path(logfile_name)
        if len(op_in["outputfiles"]) == 0:
            for ifile in glob.glob("*"):
                if ifile.startswith("."):
                    continue 
                outpath.append(Path(ifile))
        else:
            for i in op_in["outputfiles"]:
                for j in glob.glob(i):
                    if os.path.exists(j):
                        outpath.append(Path(j))
                    else:
                        log += "\n%s is not exist" % j
            outpath.append(logfile)
            if cpuinfo_log:
                outpath.append(Path(cpuinfo_log))
            if os.path.isfile("version.dat"):
                outpath.append(Path("version.dat"))
        #print("log:",log,file=sys.stderr)
        logfile.write_text(log)

        print("outpath:",str(outpath),file=sys.stderr)
        op_out = OPIO(
            {
                "outputs": outpath
            }
        )
        return op_out
    
def produce_postdft(setting,prestep_output,flowname,example_path):
    # if rundft_output is not None, then the input of postdft is the output of rundft,
    # and the example of rundft is invalid.
    # if rundft_step is None, then the input of rundft is examples
    # postdft_set is a dict
    comm.printinfo("\nPreparing post_dft")
    allsteps = []
    allstepname = []
    all_sub_save_path = []
    
    dflow_stepname = "postdft"
    bohri_stepname = flowname + f"/postdft"
    
    extrafiles, extrafiles_name = comm.transfer_source_to_artifact(
        setting.get("extra_files", []),
        source=setting.get("extra_files_source"),
        source_type=setting.get("extra_files_source_type", "local"),
        only_folder=False,
        oneartifact=True)
    
    pre_script = ""
    if setting.get("upload_datahub", None):
        pre_script += "import os\nos.system('pip install --upgrade dp-metadata-sdk==3.0.1 -i https://repo.mlops.dp.tech/repository/pypi-group/simple')\n"
    if setting.get("upload_tracking", None):
        pre_script += "import os\nos.system('pip install dp-tracking-sdk==3.15.2.post23 -i https://repo.mlops.dp.tech/repository/pypi-group/simple')\n"
        
    if not prestep_output:
        assert example_path or setting.get("example") != None, "example is not defined in postdft"
        if "example" not in setting:
            example_list = example_path
            example_source = None
            example_source_type = "local"
        else:
            example_list = setting["example"]
            example_source = setting.get("example_source")
            example_source_type = setting.get("example_source_type", "local")
        examples, examples_name = comm.transfer_source_to_artifact(
            example_list,
            source=example_source,
            source_type=example_source_type,
            only_folder=False,
            oneartifact=True)
        if examples:
            artifact_example = examples[0][0]
        else:
            comm.printinfo("No example for post-dft")
            return None,None,None
        space = "\n" + (len(bohri_stepname)+2)*" "
        comm.printinfo("%s: %s" % (bohri_stepname,space.join(examples_name[0])))
        example_path_list = examples_name[0]
    else:
        artifact_example = prestep_output
        example_path_list = example_path     
        
    executor, bohrium_set = comm.ProduceExecutor(setting, group_name=bohri_stepname)
    image = globV.get_value("ABBREVIATION").get(setting.get("image"), setting.get("image"))
    parameters = {
        "command": setting.get("command", ""),
        "outputfiles": setting.get("outputs", []),
        "metrics": setting.get("metrics", []),
        "super_metrics": setting.get("super_metrics", {}),
        "upload_tracking": setting.get("upload_tracking", {})
    }
    
    pt = PythonOPTemplate(PostDFT,image=image,envs=comm.SetEnvs())
    artifacts={"examples": artifact_example }
    if extrafiles:
        artifacts["extra_files"]=extrafiles[0][0]
    step = Step(name=dflow_stepname, template=pt,
                parameters=parameters,
                artifacts=artifacts,
                key=dflow_stepname
                )
    if executor != None:
        step.executor = executor

    allstepname.append(dflow_stepname)
    all_sub_save_path.append("")
    allsteps.append(step)
    comm.printinfo("image: %s" % image)
    comm.printinfo("set bohrium: %s" % str(bohrium_set))
    comm.printinfo("command: %s" % str(setting.get("command")))
    return allsteps, allstepname, all_sub_save_path



import traceback
from dp.launching.typing.basic import BaseModel
import os

from dp.launching.report import Report,AutoReportElement,ReportSection,ChartReportElement

from . import (comm_class,
               comm_func,
               comm_pmetrics,
               comm_class_exampleSource,
               comm_class_prepare,
               comm_class_predft,
               comm_class_rundft,
               comm_class_postdft,
               comm_class_metrics,
               readsetting)

class DatasetsModel(
    comm_class.TrackingSet,
    comm_class_metrics.metricsSaveFileSet,
    comm_class_metrics.MetricsSet,
    comm_class_postdft.PostdftImageSet,
    comm_class_postdft.PostdftCommandSet,
    comm_class_exampleSource.PostdftExtraFileNeededSet,
    comm_class_rundft.RundftImageSet,
    comm_class_rundft.RundftCommandSet,
    comm_class_rundft.RundftGroupSizeSet,
    comm_class_exampleSource.RundftExtraFileNeededSet,
    comm_class_predft.PredftImageSet,
    comm_class_predft.PredftCommandSet,
    comm_class_exampleSource.PredftExtraFileNeededSet,
    comm_class_prepare.PrepareSet,
    comm_class_prepare.PrepareOrbLibPathSet,
    comm_class_prepare.PreparePPLibPathSet,
    comm_class_prepare.PrepareDPKSDescriptorPathSet,
    comm_class_prepare.PrepareKptTemplatePathSet,
    comm_class_prepare.PrepareStruTemplatePathSet,
    comm_class_prepare.PrepareInputTemplatePathSet,
    comm_class_exampleSource.PrepareExtraFileNeededSet,
    comm_class_exampleSource.PrepareExampleSet,
    comm_class_exampleSource.DatasetSet,
    comm_class.OutputSet,
    comm_class.ConfigSet,
    BaseModel):
    ...      

def DatasetsModelRunner(opts: DatasetsModel) -> int:
    try:
        logs = comm_class.myLog()

        paths = comm_func.create_path(str(opts.IO_output_path))
        output_path = paths["output_path"]
        work_path = paths["work_path"]
        download_path = paths["download_path"]

        allparams = readsetting.ReadSetting(logs,opts,work_path,download_path)
        if allparams == None:
            return 1

        stdout,stderr = comm_func.exec_abacustest(allparams,work_path)
        logs.iprint(f"{stdout}\n{stderr}\nrun abacustest over!\n")
        reports = comm_pmetrics.produce_metrics_superMetrics_reports(allparams,work_path,output_path)

        logfname = "output.log"
        logs.write(os.path.join(str(opts.IO_output_path),logfname))
        log_section = ReportSection(title="",
                                  elements=[AutoReportElement(title='', path=logfname, description="")])
        reports.append(log_section)

        if reports:
            report = Report(title="abacus test report",
                            sections=reports,
                            description="a report of abacustest")
            report.save(output_path)
            
        #move results to output_path
        comm_func.move_results_to_output(work_path,output_path,allparams.get("save_path","results"))
    except:
        traceback.print_exc()
        return 1

    return 0


"""
MODEL_ARGS - Model registry for abacustest

This dictionary defines all available models in abacustest.
Each model is identified by a unique key (model_name) that is used as the subcommand.

Format of each entry:
    "model_name": {
        "description": "Brief description of the model (shown in help)",
        "file": "Filename of the model module (without .py extension)",
        "class_name": "Name of the Model subclass in the module"
    }

Adding a new model:
1. Create a new file in abacustest/lib_model/ (e.g., model_xxx_MyModel.py)
2. Define a class that inherits from Model (from .model import Model)
3. Implement the required methods:
   - add_args(parser): Add command-line arguments
   - run(params) / run_prepare(params) / run_postprocess(params): Implement functionality
   - Optional: HAS_PREPARE_POST_COMMAND = False if model doesn't have prepare/post subcommands
4. Add an entry to MODEL_ARGS with:
   - Key: The model name (used as subcommand)
   - description: Brief description
   - file: Module filename (without .py)
   - class_name: The class name (must match the class definition)

Note: 
- The model_name in the key must be unique and should be lowercase with hyphens if needed
- The file should follow the naming pattern: model_###_Name.py (e.g., model_022_newmodel.py)
- The class_name must exactly match the class defined in the module
- No need to implement model_name() or description() methods in the class - they are defined here
"""

MODEL_ARGS = {
    "committest": {
        "description": "Prepare the test on different abacus commit",
        "file": "model_000_CommitTest",
        "class_name": "CommitTest"
    },
    "conv": {
        "description": "Do a convergence test",
        "file": "model_001_Conv",
        "class_name": "ConvEcutwfc"
    },
    "aserelax": {
        "description": "Prepare the inputs for relax by ASE + ABACUS",
        "file": "model_003_AseRelax",
        "class_name": "AseRelax"
    },
    "eos": {
        "description": "Prepare and postprocess the EOS calculation",
        "file": "model_004_Eos",
        "class_name": "Eos"
    },
    "phonon": {
        "description": "Prepare and postprocess the phonon calculation",
        "file": "model_005_Phonon",
        "class_name": "Phonon"
    },
    "fdforce": {
        "description": "finite difference of force",
        "file": "model_006_FDForce",
        "class_name": "fdforce"
    },
    "fdstress": {
        "description": "finite difference of stress",
        "file": "model_007_FDStress",
        "class_name": "fdstress"
    },
    "comparem": {
        "description": "Compare two metrics.json files",
        "file": "model_008_CompareMetrics",
        "class_name": "CompareMetricsModel"
    },
    "fdmagforce": {
        "description": "finite difference of magenetic force",
        "file": "model_009_FDMagForce",
        "class_name": "fdmagforce"
    },
    "magj": {
        "description": "Calculate the magnetic exchange interactions of two atom",
        "file": "model_010_MagJ",
        "class_name": "magj"
    },
    "sptest": {
        "description": "Show the results of FP test",
        "file": "model_011_sptest",
        "class_name": "SPTestModel"
    },
    "band": {
        "description": "Prepare and postprocess the calcualtion of band structure",
        "file": "model_012_band",
        "class_name": "BandModel"
    },
    "inputs": {
        "description": "Prepare the ABACUS inputs of specified model",
        "file": "model_013_inputs",
        "class_name": "InputsModel"
    },
    "vasp2abacus": {
        "description": "Transform VASP input files to ABACUS input files.",
        "file": "model_014_vasp2abacus",
        "class_name": "Vasp2AbacusModel"
    },
    "elastic": {
        "description": "Prepare and postprocess the elastic",
        "file": "model_015_elastic",
        "class_name": "ElasticModel"
    },
    "bec": {
        "description": "Calculate the Born effective charge by finite difference method.",
        "file": "model_016_bec",
        "class_name": "BECModel"
    },
    "vacancy": {
        "description": "Calculate the vacancy formation energy for uncharged systems",
        "file": "model_017_vacancy",
        "class_name": "VacancyModel"
    },
    "supercell": {
        "description": "extend the unit cell to supercell",
        "file": "model_018_supercell",
        "class_name": "SuperCellModel"
    },
    "vibration": {
        "description": "Calculate the vibration frequency of selected atoms",
        "file": "model_019_vibration",
        "class_name": "VibrationModel"
    },
    "workfunc": {
        "description": "Prepare and postprocess the work function calculation",
        "file": "model_020_workfunc",
        "class_name": "WorkFuncModel"
    },
    "dos-pdos": {
        "description": "Postprocess the DOS and PDOS data from ABACUS calculations",
        "file": "model_021_dos_pdos",
        "class_name": "DOSPDOSModel"
    }
}


from dp.launching.typing.basic import BaseModel, Int, String, Float,List,Optional,Union,Dict
from dp.launching.typing import InputFilePath, OutputDirectory
from dp.launching.typing import (
    Boolean,
    Field
)
from dp.launching.report import Report

from . import comm_class,comm_func
import json,traceback,os,shutil,datetime,zipfile
from dp.metadata import MetadataContext, Dataset
from dp.metadata.utils.storage import TiefblueStorageClient
from abacustest.myflow.comm import CollectFileName
from dp.launching.report import ChartReportElement,ReportSection,AutoReportElement

datahub_project = "abacustest_dataset"

io_input_path_description = f"""A compressed file contains all example folders. 

For each folder is one example, and containing all the required files.

The dataset will be uploaded to https://datahub.mlops.dp.tech/browse/dataset/corp/tiefblue/{datahub_project}
"""

class UplaodDataset(BaseModel):
    IO_input_path:InputFilePath = Field(title="Upload dataset",
                                        st_kwargs_type = comm_func.unpack(None,None,get_support_filetype=True), 
                                        description=io_input_path_description,
                                        description_type="markdown")
    IO_output_path: OutputDirectory = Field(default="./output")
    name: String = Field(title="Dataset Name",regex="^\\s*[a-zA-Z0-9_-.]+\\s*$",description="Can only contains letters, numbers, dot(.), _ and -. (regex is: [a-zA-Z0-9_-.])")
    overwrite: Boolean = Field(description="If overwrite when the dataset already exists? Only owner can overwrite it.")
    description: String = Field(default="")

class UplaodDatasetModel(UplaodDataset,comm_class.ConfigSet,BaseModel):
    ...

def upload(download_path, bohrium_username, bohrium_password, bohrium_project, dataset_name, new_dataset_name_tail, tags, properties, description, overwrite, logs):
    '''
    dataset.description   dataset.entity_type   dataset.group         dataset.properties    dataset.uri           
    dataset.display_name  dataset.gen_urn(      dataset.owners        dataset.tags          dataset.urn
    '''
    cwd = os.getcwd()
    os.chdir(download_path)
    mess = description + "\\\n" + "\\\n".join(CollectFileName("."))
    os.chdir(cwd)

    metadata_storage_client = TiefblueStorageClient(
        bohrium_username, bohrium_password, bohrium_project)
    with MetadataContext(storage_client=metadata_storage_client, project=datahub_project) as context:
        def produce_unique_urn(dataset_name, bohrium_username=None, overwrite=False):
            org_name = dataset_name
            n = 1
            while True:
                urn = Dataset.gen_urn(context, platform="tiefblue",
                                      name=dataset_name, auto_suffix=False)
                dataset = client.get_dataset(urn)
                if (dataset == None) or ():
                    return dataset_name, urn
                elif overwrite and bohrium_username in dataset.tags:
                    logs.iprint(
                        f"\nDataset '{dataset_name}' already exists!\n\turn: {urn}\n\ttags: {dataset.tags}\nOverwrite it!")
                    return dataset_name, urn

                logs.iprint(
                    f"\nDataset '{dataset_name}' already exists!\n\turn: {urn}\n\ttags: {dataset.tags}")
                dataset_name = org_name + f"_{n}"
                n += 1
                logs.iprint(f"Rename dataset to '{dataset_name}'")

        client = context.client
        urn = Dataset.gen_urn(context, platform="tiefblue",
                              name=dataset_name, auto_suffix=False)
        old_dataset = client.get_dataset(urn)
        if old_dataset != None and (bohrium_username not in old_dataset.tags) and (not dataset_name.endswith(new_dataset_name_tail)):
            logs.iprint(
                f"\nDataset '{dataset_name}' already exists!\n\turn: {urn}\n\ttags: {old_dataset.tags}")
            dataset_name += new_dataset_name_tail
            logs.iprint(f"Rename the dataset to '{dataset_name}'")
        if overwrite:
            dataset_name, urn = produce_unique_urn(
                dataset_name, bohrium_username, True)
        else:
            dataset_name, urn = produce_unique_urn(dataset_name)

        uri = client.upload_artifact(None, None, download_path)
        dataset = Dataset(
            urn=urn,
            uri=uri,
            display_name=dataset_name,
            tags=tags,
            description=mess,
            properties=properties
        )
        urn1 = client.create_dataset(dataset)
        dataset = client.get_dataset(urn1)
        if dataset == None:
            logs.iprint("\nUpload to datahub failed!\nPlease contact developer to deal with it.")
            return False
        else:
            website = f"https://datahub.mlops.dp.tech/dataset/{urn1}/Documentation?is_lineage_mode=false"
            logs.iprint(f"\nUpload dataset to datahub successfully.")
            logs.iprint(f"\tname: {dataset.display_name}")
            logs.iprint(f"\turn : {dataset.urn}")
            logs.iprint(f"\ttags: {dataset.tags}")
            logs.iprint(f"\twebsite: {website}")
            return True

def UplaodDatasetModelRunner(opts:UplaodDatasetModel):
    download_path = "download_data"
    if os.path.isdir(download_path):
        shutil.rmtree(download_path)
    os.makedirs(download_path,exist_ok=True)
    os.makedirs(str(opts.IO_output_path),exist_ok=True)

    with zipfile.ZipFile(opts.IO_input_path.get_path(), "r") as zip_ref:
        zip_ref.extractall(download_path)

    today = datetime.datetime.now().strftime('%Y/%m/%d-%H:%M')
    new_dataset_name_tail = "_" + opts.Config_bohrium_username.split("@")[0]
    overwrite = opts.overwrite
    tags = [opts.Config_bohrium_username,opts.Config_bohrium_project_id,today.split("-")[0],today]
    properties = {"bohrium_project":opts.Config_bohrium_project_id}

    logs = comm_class.myLog()
    doupload = upload(download_path,
           opts.Config_bohrium_username, 
           opts.Config_bohrium_password, 
           opts.Config_bohrium_project_id, 
           opts.name.strip(), 
           new_dataset_name_tail, 
           tags, 
           properties, 
           opts.description,
           overwrite, 
           logs)

    logfname = "output.log"
    logs.write(os.path.join(str(opts.IO_output_path),logfname))

    report = Report(title="",
                    sections=[ReportSection(title="",
                              elements=[AutoReportElement(title='', path=logfname, description="")])],
                    description="")
    report.save(str(opts.IO_output_path))
    if doupload:
        return 0
    else:
        return 1







#!/usr/bin/env python3
import os,sys,glob,time,shutil,argparse,json,traceback,re
import numpy as np

from dflow import (
    Workflow,
    download_artifact,
)

import  os, shutil, glob
from . import globV,dflowOP,comm

def ParamParser(param):
    """
    {
        "bohrium_goup_name": ,
        "ABBREVIATION":{},
        "save_path" : PATH_THE_FINAL_RESULT_WILL_BE_DOWNLOADED_TO,
        "pre_dft":{},
        "run_dft" : {},
        "post_dft": {},
        "report":{},
        "remote": {},
        "compress": true, # if compress the inputs and outputs
    }
    """
    
    alljobs = {}
    alljobs["compress"] = param.get("compress",True) # if use slice, we can compress the inputs
    alljobs["save_path"] = param.get("save_path",None)
    alljobs["prepare"] = param.get("prepare",{"ifrun":False})
    alljobs["pre_dft"] = param.get("pre_dft",{"ifrun":False})
    alljobs["run_dft"] = param.get("run_dft",{"ifrun":False})
    alljobs["post_dft"] = param.get("post_dft",{"ifrun":False})
#    alljobs["upload_datahub"] = param.get("upload_datahub",None)  #used to upload local files to datahub
#    alljobs["upload_tracking"] = param.get("upload_tracking",None)  #used to upload tracking
    alljobs["report"] = param.get("report",None)
    alljobs["remote"] = param.get("remote",None)
    alljobs["bohrium_group_name"] = "abacustesting"
    if param.get("bohrium_group_name","") != "":
        alljobs["bohrium_group_name"] = param.get("bohrium_group_name")
    elif param.get("config",{}).get("dflow_labels",{}).get("launching-job"):
        bohriu_name = param.get("config",{}).get("dflow_labels",{}).get("launching-job")
        if bohriu_name.startswith("sched-abacustest-"):
            alljobs["bohrium_group_name"] = "s-" + bohriu_name[17:]
        elif bohriu_name.startswith("job-abacustest-"):
            alljobs["bohrium_group_name"] = "j-" + bohriu_name[15:]
        else:
            alljobs["bohrium_group_name"] = bohriu_name
        
    alljobs["ABBREVIATION"] = param.get("ABBREVIATION",{})

    #print(alljobs)
    #sys.exit(1)
    globV.set_value("ABBREVIATION",alljobs.get('ABBREVIATION',{}))
    return alljobs

def SetSaveFolder(storefolder=None):
    if storefolder == None:
        #import datetime
        #from time import strftime
        #today = datetime.datetime.now()
        #today = today.strftime("%Y%m%d")
        #storefolder = os.path.join("result",today)
        storefolder = "results"
    globV.set_value("RESULT",storefolder)
    comm.printinfo("set save folder: %s" % storefolder)

def MakeSaveFolder(storefolder=None):
    storefolder = globV.get_value("RESULT") if storefolder == None else storefolder
    storefolder = storefolder.strip().strip("/")
    if not os.path.isdir(storefolder):
        os.makedirs(storefolder)
    elif not globV.get_value("OVERRIDE"):
        n = 1
        bk = storefolder + ".bk%d" % n
        while os.path.isdir(bk):
            n += 1
            bk = storefolder + ".bk%d" % n
        os.rename(storefolder,bk)
        comm.printinfo("Folder %s is exist, rename to %s" % (storefolder,bk))
        os.makedirs(storefolder)
    
def WriteParamUserFile(storefolder=None,override=False):
    storefolder = globV.get_value("RESULT") if storefolder == None else storefolder
    paraf = os.path.join(storefolder,globV.get_value("PARAM_FNAME")) 
    
    if not override:
        def getfname(paraf):
            paraf_tmp = paraf
            n = 1
            while os.path.isfile(paraf_tmp):
                paraf_tmp = paraf + "%d" % n
                n += 1
            return paraf_tmp
            
        paraf = getfname(paraf)

    save_cotext = {}
    for k,v in globV.get_value("PARAM_CONTEXT").items():
        if k in ["config"]:
            continue
        
        # hide the config information in dispatcher
        if isinstance(v,dict):
            if "dispatcher" in v:
                comm.hide_config_in_dispatcher(v["dispatcher"])
        elif isinstance(v,list):
            for i in range(len(v)):
                if isinstance(v[i],dict):
                    if "dispatcher" in v[i]:
                        comm.hide_config_in_dispatcher(v[i]["dispatcher"])
        
        # hide remote/github/token information
        if k == "remote":
            if "github" in v and "token" in v["github"] and v["github"]["token"].strip() != "":
                v["github"]["token"] = "******"
        save_cotext[k] = v
    json.dump(save_cotext,open(paraf,'w'),indent=4)
    #with open(paraf,'w') as f1: f1.write(globV.get_value("PARAM_CONTEXT")) 

def set_config(param_context,debug):
    # read config from param.json and os.environ
    # key in param.json should be lower case, and in os.environ should be upper case
    # support older key name, such as "bohrium_username" and "lbg_username"
    '''
    old keys:   "lbg_username","lbg_password","bohrium_ticket","project_id",
                 "config_host","s3_config_endpoint","config_k8s_api_server","config_token",
                 "datahub_project","datahub_gms_token","datahub_gms_url","AIM_ACCESS_TOKEN",
                 "dflow_labels"
    datahub is not used anymore.
    '''
    
    if "config" in param_context:
        user_context = param_context.get("config")
    else:
        comm.printinfo(f"WARNING: \"config\" is not detected in parameter file, try to read config information from os.environ.")
        user_context = {}
    
    configs = {} 
    for new_key,old_key in [["bohrium_username","lbg_username"],
                            ["bohrium_password","lbg_password"],
                            ["bohrium_ticket","bohrium_ticket"],
                            ["bohrium_project_id","project_id"],
                            ["dflow_host","config_host"],
                            ["dflow_s3_config_endpoint","s3_config_endpoint"],
                            ["dflow_k8s_api_server","config_k8s_api_server"],
                            ["dflow_token","config_token"],
                            ["dflow_labels","dflow_labels"],   # dflow_labels is a dict
                            ["aim_access_token","AIM_ACCESS_TOKEN"],
                            ["github_username","github_username"],
                            ["github_password","github_password"],
                            ["github_token","github_token"]
    ]:
        if new_key in user_context:
            configs[new_key] = user_context.pop(new_key)
        elif old_key in user_context:
            configs[new_key] = user_context.pop(old_key)
        elif new_key.upper() in os.environ:
            configs[new_key] = os.environ[new_key.upper()]
        elif old_key.upper() in os.environ:
            configs[new_key] = os.environ[old_key.upper()]
    
    if "dflow_labels" in configs and isinstance(configs["dflow_labels"],str):
        if configs["dflow_labels"].strip() != "":
            import ast
            print("dflow_labels:",configs["dflow_labels"].strip())
            configs["dflow_labels"] = ast.literal_eval(configs["dflow_labels"].strip())
            print("dflow_labels literal_eval:",configs["dflow_labels"])
            
        else:
            del configs["dflow_labels"]
    
    for ik,iv in user_context.items():
        configs[ik] = iv
            
    globV.set_value("PRIVATE_SET", configs)
    
    if configs.get("dflow_labels",None) != None:
        if isinstance(configs["dflow_labels"],dict) and configs["dflow_labels"].get("launching-job"):
            job_address = "https://app.bohrium.dp.tech/abacustest/?request=GET%3A%2Fapplications%2Fabacustest%2Fjobs%2F" + configs["dflow_labels"].get("launching-job")
            globV.set_value("JOB_ADDRESS",job_address)
    
    dflowOP.SetConfig(configs,debug=debug)
    return 
    
def set_env(param):
    globV.set_value("OVERRIDE", param.override)
    comm.printinfo("\nSet enviroment ...")
    comm.printinfo(param)

    #read job.json
    if not os.path.isfile(param.param):
        comm.printinfo("ERROR: Can not find the test setting file '%s' " % param.param)
        sys.exit(1)
    comm.printinfo("Read parameter setting from %s" % param.param)
    globV.set_value("PARAM_FNAME", os.path.split(param.param)[1]) 
    globV.set_value("PARAM_CONTEXT", json.load(open(param.param)))       
    param_context = json.load(open(param.param))
    globV.set_value("PARAM", param_context)
    
    #read user config information
    bohrium_executor = False
    if "bohrium_executor" in param_context:
        bohrium_executor = bool(param_context["bohrium_executor"])
    globV.set_value("BOHRIUM_EXECUTOR",bohrium_executor)
    
    set_config(param_context,param.debug)

    #set save folder  
    if param_context.get("save_path"):
        save_path = param_context.get("save_path")
    else:
        save_path = param.save
    SetSaveFolder(save_path)

    report = param_context.get("report",{})
    if "report" in param_context and report.get("ifrun",True):
        globV.set_value("REPORT", report)
        
    remote = param_context.get("remote",{})
    if "remote" in param_context and remote.get("ifrun",True):
        globV.set_value("REMOTE", remote)
        
    compress = "default" if param_context.get("compress",True) else None
    globV.set_value("COMPRESS",compress)
    #if compress:
    #    comm.printinfo("Compress the inputs to upload.")


def waitrun(wf,stepnames,allsave_path):
    '''
    stepnames = [[test1_stepname1,test1_stepname2,...],[test2_stepname1,test2_stepname2,...],...]
    allsave_path = [[[save_path,sub_save_path],[save_path,sub_save_path],...],[]...] similar to stepnames
    postdft_local_jobs = [[],[save_path,param["post_dft"]],[],..], null list means no postdft_local,
    '''
    finishtest = []
    makedfolder = []
    for i,istep in enumerate(stepnames):
        finishtest.append(len(istep)*[False])
    finishtest = np.array(finishtest) 
          
    wfid = wf.id

    while False in finishtest:
        for i,istep in enumerate(stepnames):
            for j,jfinish in enumerate(finishtest[i]):
                if jfinish:
                    continue
                steps = wf.query_step(name = istep[j])
                if len(steps) > 0: 
                    allfinished = True
                    for step in steps:
                        #print(len(steps),step.name,step.phase)
                        if step.phase in ["Pending","Running"]:
                            allfinished = False
                            time.sleep(4)
                            break
                    if not allfinished:
                        continue
                    step = steps[0]
                    finishtest[i][j] = True

                    save_path = allsave_path[i][j][0]
                    part_save_path = os.path.join(allsave_path[i][j][0],allsave_path[i][j][1])                      
                        
                    if part_save_path not in makedfolder:
                        MakeSaveFolder(part_save_path)
                        WriteParamUserFile(storefolder=part_save_path)
                        makedfolder.append(part_save_path)
                    comm.printinfo("%s is finished (remaining %d jobs for this test), download the results to %s!" % 
                                   (step.name,len(finishtest[i]) - np.sum(finishtest[i]),part_save_path))
                    if step.phase != 'Succeeded':
                        comm.printinfo("    This job is not Succeeded, please check on: %s, workflow ID is: %s" %
                              (globV.get_value("HOST"),wfid))  
                    try:
                        print("artifacts outputs:",step.outputs.artifacts["outputs"])
                        print("step.outputs:",step.outputs)
                        download_artifact(step.outputs.artifacts["outputs"],path=save_path)
                    except:
                        traceback.print_exc()
            if wf.query_status().strip() == "Failed":
                job_address = globV.get_value("HOST") + "/%s?tab=workflow" % wf.id
                comm.printinfo(f"The workflow is failed, please check on: {job_address}")  
                return
                    
                                
        time.sleep(4)

def ReportMetrics():
    report_setting = globV.get_value("report",{})
    if not report_setting or not report_setting.get("ifrun",True):
        return

    from abacustest import outresult
    allresults = outresult.GetAllResults(report_setting)
    if not allresults:
        return
    
    split_example=None if len(allresults["type_name"]) == 1 else "----"
    cc_outparam,allparam_value = outresult.OutParam(allresults,split_example=split_example)
    cc_outmetrics,allmetric_value = outresult.OutMetrics(allresults,allparam_value)
    
    #report
    from datetime import datetime
    param_context = globV.get_value("PARAM")
    report  = """\n\n\t\tABACUS TESTING REPORT\n"""
    report += "testing begin: %s\n" % globV.get_value("START_TIME").strftime("%d/%m/%Y %H:%M:%S")
    report += "testing   end: %s\n" % str(datetime.now().strftime("%d/%m/%Y %H:%M:%S"))
    report += "run_dft setting:\n"
    for irun in param_context.get("run_dft",[]):
        image = globV.get_value("ABBREVIATION").get(irun.get("image"),irun.get("image"))
        bohrium = irun.get("bohrium")
        report += "\timage: %s\n\tbohrium:%s\n\texample:%s\n" % (image,str(bohrium),str(irun.get("example")))
        
    if len(allresults["metrics"]) > 0:
        report += cc_outmetrics
    report += cc_outparam

    report_file = report_setting.get("save_file")
    if report_file != None:
        with open(report_file,'w') as f1:
            f1.write(report)
    comm.printinfo(report)

def upload_packages_(packages):
    if packages is not None:
        from dflow.python import upload_packages
        for package in packages:
            if not os.path.isdir(package):
                raise Exception(f"upload_package {package} is not a valid folder!")
            if package not in upload_packages:
                upload_packages.append(package)

def RunJobs(param):
    upload_packages_(json.load(open(param.param)).get("upload_packages",None))
    set_env(param)
    alljobs = ParamParser(json.load(open(param.param)))
    allstep,stepname,allsave_path = dflowOP.ProduceAllStep(alljobs)

    job_id = None
    if len(allstep) == 0:
        comm.printinfo("No step is produced, exit!!!")
    else:
        dflow_labels = globV.get_value("PRIVATE_SET",{}).get("dflow_labels",{})
        if "launching-application" not in dflow_labels:
            dflow_labels["launching-application"] = "abacustest"
        max_para = globV.get_value("PARAM").get("max_parallel",100)
        if globV.get_value("BOHRIUM_EXECUTOR"):
            wf = Workflow(name="abacustest",context=globV.get_value("BRM_CONTEXT"),labels=dflow_labels,parallelism=max_para)
        else:
            wf = Workflow(name="abacustest",labels=dflow_labels,parallelism=max_para)

        wf.add(allstep)
        wf.submit()
        
        job_id = wf.id
        comm.printinfo("job ID: %s, UID: %s" % (job_id,wf.uid))
        job_address = globV.get_value("HOST") + "/%s?tab=workflow" % wf.id
        comm.printinfo("You can track the flow by using your browser to access the URL:\n %s\n" % job_address)

        if not param.download:
            return job_id
        waitrun(wf,stepname,allsave_path)
    
    #if globV.get_value("REPORT"):
    #    ReportMetrics()
    
    if globV.get_value("REPORT"):
        comm.printinfo("\nGenerate html report ...")
        pwd = os.getcwd()
        if os.path.isdir(globV.get_value("RESULT")):
            os.chdir(globV.get_value("RESULT"))

        from abacustest import report
        filename = "abacustest.html"
        report.gen_html(globV.get_value("REPORT"),filename)
        
        os.chdir(pwd)
    
    if globV.get_value("REMOTE"):
        comm.printinfo("\nPush the results to remote ...")
        pwd = os.getcwd()
        if os.path.isdir(globV.get_value("RESULT")):
            os.chdir(globV.get_value("RESULT"))
        
        from abacustest import remote
        remote_setting = globV.get_value("REMOTE")
        if "github" in remote_setting and "token_file" in remote_setting["github"] and remote_setting["github"]["token_file"].strip() != "":
            remote_setting["github"]["token"] = open(os.path.join(pwd,remote_setting["github"]["token_file"].strip())).read().strip()
        remote_path = remote.prepare_files(remote_setting)
        if remote_path:
            remote.push_to_remote(remote_path,remote_setting,globV.get_value("PRIVATE_SET",{}))
        
        os.chdir(pwd)
    
    return job_id
        
def CheckStatus(param):
    if os.path.isfile(param.param):
        private_set = json.load(open(param.param))
        if "config" in private_set:
            private_set = private_set["config"]
        elif "USER" in private_set:
            private_set = private_set["USER"]
    else:
        print("config file is not found!\nUse the default setting!")
        private_set = {}
        
    set_config(private_set,False) 
    
    jobid = param.job_id
    
    wf = Workflow(id = jobid)
        
    try:
        return wf.query_status()
    except:
        #comm.printinfo("Query status error")
        #traceback.print_exc()
        return "The flow is not running now!!!"
        
def ReadUIDFromLog(logfile):
    with open(logfile) as f1: lines = f1.readlines()
    
    uid = None
    job_id = None
    for line in lines[::-1]:
        if line.startswith("job ID:"):
            "job ID: abacustest-r7gw6, UID: 10930511-04c2-4cb1-ba00-d55f72b3270f"
            print("Find job ID line:", line)
            try:
                job_id = line.split(",")[0].split(":")[1].strip()
                uid = line.split(",")[1].split(":")[1].strip()
            except:
                return None, None
            break
            
    return job_id, uid



def DownloadFlow(param):
    if os.path.isfile(param.param):
        private_set = json.load(open(param.param))
        if "USER" in private_set:
            private_set = {"config":private_set["USER"]}
    else:
        print("config file is not found!\nUse the default setting!")
        private_set = {}
    set_config(private_set,False)   
    
    save_path = comm.ParseSavePath(private_set.get("save_path",None)) 
    if param.save is None:
        save_path = comm.ParseSavePath(private_set.get("save_path","result"))
    else:
        save_path = param.save
    
    jobid = uid =  param.job_id
    if jobid is None and os.path.isfile("abacustest.log"):
        print("Try to find job ID from log file!!!")
        jobid, uid = ReadUIDFromLog("abacustest.log")
        if jobid is None:
            comm.printinfo("Can not find job ID from log file!!!")
            return
        
    try:
        wf = Workflow(id = jobid)
        allkeys = wf.query_keys_of_steps()
    except:
        comm.printinfo(f"Query keys with ID: {jobid} failed!!!")
        if uid != jobid:
            comm.printinfo(f"Try UID: {uid}")
            wf = Workflow(id = uid)
            allkeys = wf.query_keys_of_steps()
        else:
            traceback.print_exc()
            return
    
    if len(allkeys) == 0:
        comm.printinfo("No step is produced, exit!!!")
        return
    
    final_step = allkeys[-1].split("-")[0] # key: rundft-1-1,.. (rundft use slice, so the key is rundft-1-1,..)
    if final_step in ["predft", "rundft", "postdft"]:
        download_steps = [i for i in wf.query_step() if i.key is not None and i.key.split("-")[0] == final_step]
        # because in rundft step, it use slice, so each rundft group only download once is ok.
        # then based on the step name, seperate the steps to different group
        download_groups = {}
        
        for step in download_steps:
            # the group_name is like: abacustest-7589t[0].abacustesting[1].rundft-2(1:1)
            group_name = step.name.split(".")[-1].split("(")[0]
            if group_name not in download_groups:
                download_groups[group_name] = []
            download_groups[group_name].append(step)
        
        for group_name,group_steps in download_groups.items():
            phases = [i.phase for i in group_steps]
            if not all([bool(i in ["Succeeded"]) for i in phases]):
                comm.printinfo(f"Below steps are not Succeeded:")
                for step in group_steps:
                    if step.phase not in ["Succeeded"]:
                        comm.printinfo(f"    phase of \"{step.name}\" (key: \"{step.key}\") is {step.phase}")
            download_artifact(group_steps[0].outputs.artifacts["outputs"],path=save_path)
    else:
        comm.printinfo("The final step is not predft, rundft or postdft, can not download the results!")
        



import argparse, importlib, os, sys
from abacustest.lib_model.model_args import MODEL_ARGS


def _import_model_class(model_info):
    """Import model class from MODEL_ARGS info, with error handling."""
    try:
        module = importlib.import_module(f"abacustest.lib_model.{model_info['file']}")
        return getattr(module, model_info['class_name'])
    except ImportError as e:
        print(f"Error: Failed to import model '{model_info['file']}'")
        print(f"Import error: {e}")
        print("This may be due to missing dependencies. Please install the required packages.")
        sys.exit(1)
    except AttributeError as e:
        print(f"Error: Failed to find class '{model_info['class_name']}' in model '{model_info['file']}'")
        print(f"Attribute error: {e}")
        sys.exit(1)


def ModelArgs(parser, list_all_models=False, model_subcommand=None):  
    subparser = parser.add_subparsers(dest="model")
    
    if list_all_models:
        model_names = MODEL_ARGS.keys()
        
        for model_name in model_names:
            model_info = MODEL_ARGS[model_name]
            iparser = subparser.add_parser(model_name, help=model_info["description"])
            iparser.description = model_info["description"]
    else:
        if model_subcommand is None:
            return
        
        if model_subcommand not in MODEL_ARGS:
            print(f"Error: unknown model '{model_subcommand}'")
            print(f"Available models: {', '.join(sorted(MODEL_ARGS.keys()))}")
            sys.exit(1)
        
        model_info = MODEL_ARGS[model_subcommand]
        model_class = _import_model_class(model_info)
        
        iparser = subparser.add_parser(model_subcommand, help=model_info["description"])
        iparser.description = model_info["description"]
        
        if model_class.HAS_PREPARE_POST_COMMAND:
            subsubparser = iparser.add_subparsers(dest="modelcommand")
            iparser_prepare = subsubparser.add_parser("prepare", help="Prepare the model")
            iparser_postprocess = subsubparser.add_parser("post", help = "Post-process the model")
            model_class.prepare_args(iparser_prepare)
            model_class.postprocess_args(iparser_postprocess)

        model_class.add_args(iparser)
    
    parser.description = "Prepare and post-process the specified model"

    return parser

def RunModel(param):
    model_info = MODEL_ARGS[param.model]
    model_class = _import_model_class(model_info)
    imodel = model_class()
    
    print(f"Model: {param.model}")
    if not imodel.HAS_PREPARE_POST_COMMAND:
        imodel.run(param)
    else:
        if param.modelcommand == "prepare":
            imodel.run_prepare(param)
        elif param.modelcommand == "post":
            imodel.run_postprocess(param)
        else:
            imodel.run(param)

def main():
    parser = argparse.ArgumentParser()
    param = ModelArgs(parser).parse_args()
    RunModel(param)
    
if __name__ == "__main__":
    main()


#!/usr/bin/env python3
import os,sys,argparse
sys.path.append(os.path.split(__file__)[0])
from .myflow import flow,globV,comm

def submitflow(param):
    from dflow.python import upload_packages
    upload_packages.append(os.path.split(__file__)[0])
    globV._init()
    flow.RunJobs(param)
    comm.printinfo("\nAll jobs are finished!!!")

def checkstatus(param):
    globV._init()
    status = flow.CheckStatus(param)
    print(status)

def downloadflow(param):
    globV._init()
    flow.DownloadFlow(param)

def main():
    from abacustest.arguments import SubmitArgs,CheckStatusArgs

    parser = argparse.ArgumentParser(description="abacustest")
    subparser = parser.add_subparsers(dest="command")
    SubmitArgs(subparser.add_parser("submit",help="Submit a workflow"))
    SubmitArgs(subparser.add_parser("mlops-submit",help="Just submit a workflow and will not wait for the result"))
    CheckStatusArgs(subparser.add_parser("status", help="Check the status of the dflow job"))
    parser = parser.parse_args()
    if parser.command == "submit":
        submitflow(parser)
    elif parser.command == "status":
        checkstatus(parser)
    elif parser.command == "download":
        downloadflow(parser)

if __name__ == '__main__':
    main()



import argparse

def _job_type(string):
    string = string.lower()
    type_map = {'0': 'abacus', '1': 'qe', '2': 'vasp'}
    if string in type_map:
        return type_map[string]
    valid_choices = list(type_map.values())
    if string not in valid_choices:
        raise argparse.ArgumentTypeError(f"Invalid job type: {string}. Valid options are: {valid_choices} or 0/1/2")
    return string

def SubmitArgs(parser):
    parser.description = "This script is used to run a testing"
    parser.add_argument('-p', '--param', type=str, default="job.json",help='the job setting file, default is job.json')
    parser.add_argument('-s', '--save', type=str, default=None,help='the folder where the results will be put in, default: result')
    parser.add_argument('--override', type=int, default=1,help="when the save folder exists, if override it. 0: no, 1: yes. ")
    parser.add_argument('--debug',nargs='?',type=int, const=1, default=0,help="debug mode for dflow")
    parser.add_argument('--download',default=1,type=int, help="if wait the finish of flow and download the results, 0: no, 1: yes. Default is 1")
    return parser

def CheckStatusArgs(parser):
    parser.description = "Check the status of the dflow job"
    parser.add_argument('-p', '--param', type=str, default="job.json",help='the file for bohrium account information, default is "job.json"')
    parser.add_argument("job_id", help="the job id of dflow")
    return parser

def DownloadArgs(parser):
    parser.description = "Download the results of the dflow job"
    parser.add_argument("job_id", default=None, nargs="?",help="the job id of your workflow")
    parser.add_argument('-p', '--param', type=str, default="job.json",help='the file for bohrium account information, default is "job.json"')
    parser.add_argument('-s', '--save', type=str, default=None,help='the folder where the results will be put in, default: result')
    return parser

def CollectDataArgs(parser):
    parser.description = "This script is used to collect some key values from the output of ABACUS/QE/VASP jobs"
    parser.add_argument('-j', '--jobs', default=["."], help='the path of jobs', action="extend",nargs="*")
    parser.add_argument('-t', '--type', type=_job_type, default='abacus', help='abacus/qe/vasp or 0/1/2. Default: abacus', choices=['abacus', 'qe', 'vasp'])
    parser.add_argument('-p', '--param', type=str, default=None,nargs="*", help='the parameter file, or parameter name.')
    parser.add_argument('-o', '--output', type=str, default="metrics.json",help='the file name to store the output results, default is "metrics.json"')
    parser.add_argument('-m', '--modules',help='add extra modules. Default only module \'job-type\' will be loaded, such as: \'abacus\' for abacus type. You can check all modules by --outparam', action="extend",nargs="*")
    parser.add_argument('--newmethods', help='the self-defined python modules, and shuold be format of import, such as "abc"(the file name is abc.py), "a.b.c" (teh file is a/b/c.py)', action="extend",nargs="*")
    parser.add_argument('--outparam', nargs='?',type=int, const=1, default=0,help='output the registed parameters, you can set the type by -t or --type to choose abacus/qe/vasp. 0: No, 1: yes')
    parser.add_argument('--ref', type=str, nargs='?',default=None,const="resultREF.json",help='A json file includes the reference value of some keys. Generally, get values of keys start with \"delta_\" require this file. Default is resultREF.json')
    return parser

def OutResultArgs(parser):  
    parser.description = "This script is used to output the summary of results"
    parser.add_argument('-r', '--result', type=str, help='the result file from collectdata, should be .json type',action="extend",nargs="*")
    parser.add_argument('-p', '--param', type=str, help='the parameter file, should be .json type')
    parser.add_argument('-m', '--metrics', default=["ALL"], help='The metrics that needed to be shown. Use ALL to show all metrics, default is ALL.', action="extend",nargs="*")
    parser.add_argument('-o', '--output', type=str, help='output the selected metrics to a json file')
    parser.add_argument('--csv', type=str, help='if output the selected metrics to csv file')
    parser.add_argument('--prec', type=int, default=None, help='the precision of float number in pandas output, default is None')
    parser.add_argument("--list_sep", default=0, const=1, nargs='?', type=int, help="if print the list separately, default is 0")
    return parser

def PrepareArgs(parser):
    parser.description = "This script is used to prepare the INPUTS OF ABACUS JOB"
    parser.add_argument('-p', '--param', type=str, help='the parameter file, should be .json type',required=True)
    parser.add_argument('-s', '--save', type=str,  default="abacustest",help='where to store the inputs, default is abacustest ')
    parser.add_argument('--nolink',  nargs='?',type=int, const=1, default=0,help='if link the files in the example folder, default is 0')
    return parser

def ReportArgs(parser):  
    parser.description = "Read metrics.json and generate the report"
    parser.add_argument('-p', '--param', type=str, help='the parameter file, should be .json type',required=True)
    parser.add_argument('-o', '--output', type=str,  default="abacustest.html",help='The output file name, default is abacustest.html')
    return parser


def RemoteArgs(parser):  
    parser.description = '''push the local files to remote, like github
    Need to specify the parameter file, which should be .json type, and the content should be like:
    "config":{
        "github_username": "",
        "github_email": "",
        "github_token": ""
    },
    "add_date": "bda/dates",
    "remote":{
        "files":{
            "file1_local": "file1_remote",
            "file2_local": "file2_remote",
        },
        "github":{
            "repo": "",
            "branch": "",
            "commit_msg": ""
        }
    }
    the key of "files" is the local file/folder, and the value is the remote file/folder.
    '''
    parser.add_argument('-p', '--param', type=str, help='the parameter file, should be .json type',required=True)
    return parser

def SkillsArgs(parser):
    parser.description = "Show the path of abacustest skills"
    return parser




from dp.launching.cli import (
    SubParser,
    default_minimal_exception_handler,
    run_sp_and_exit,
    to_runner,
    default_exception_handler
)

from abacustest.launching import (
    #model_normal,
 #model_selfDefine,
 #model_summary,
 #model_postdft,
 #model_predft,
 #model_advanced,
 model_expert,
 model_report,
 model_reuse,
 model_phonon,
 model_fdforce,
 model_fdmagforce,
 model_fdstress,
 model_autoAbacus,
 model_vasp2abacus
 )


def to_parser():
    return {
       # "Normal":SubParser(model_normal.NormalModel,model_normal.NormalModelRunner,"Simple rundft and postdft"),
       # "Advanced":SubParser(model_advanced.AdvancedModel,model_advanced.AdvancedModelRunner,"Complete rundft and postdft"),
       # "Expert":SubParser(model_expert.ExpertModel,model_expert.ExpertModelRunner,"Complete predft, rundft and postdft"),
       # "Predft":SubParser(model_predft.PredftModel,model_predft.PredftModelRunner,"Only Predft"),
       # "Postdft":SubParser(model_postdft.PostdftModel,model_postdft.PostdftModelRunner,"Only Postdft"),
       # "SettingFile":SubParser(model_selfDefine.SelfDefineModel,model_selfDefine.SelfDefineModelRunner,"run self-defined model"),
        #"NormalDatasets":SubParser(model_normal.NormalDatasetsModel,model_normal.NormalModelRunner,"Simple rundft and postdft (use launching datasets as input)"),
        #"AdvancedDatasets":SubParser(model_advanced.AdvancedDatasetsModel,model_advanced.AdvancedModelRunner,"Complete rundft and postdft (use launching datasets as input)"),
       # "ExpertDatasets":SubParser(model_expert.ExpertDatasetsModel,model_expert.ExpertModelRunner,"Complete predft, rundft and postdft (use launching datasets as input)"),
        #"PredftDatasets":SubParser(model_predft.PredftDatasetsModel,model_predft.PredftModelRunner,"Only Predft (use launching datasets as input)"),
        #"PostdftDatasets":SubParser(model_postdft.PostdftDatasetsModel,model_postdft.PostdftModelRunner,"Only Postdft (use launching datasets as input)"),
       # "SettingFileDatasets":SubParser(model_selfDefine.SelfDefineDatasetsModel,model_selfDefine.SelfDefineModelRunner,"run self-defined model(use launching datasets as input)"),
        #"UploadDatasets":SubParser(model_uploadDataset.UplaodDatasetModel,model_uploadDataset.UplaodDatasetModelRunner,"upload datasets to datahub"),
       # "Report":SubParser(model_report.ReportModel,model_report.ReportModelRunner,"report metrics.json"),
       "08-Vasp2Abacus": SubParser(model_vasp2abacus.Vasp2AbacusModel, model_vasp2abacus.Vasp2AbacusRunner, "Convert VASP jobs to ABACUS jobs"),
        "07-AutoRun":  SubParser(model_autoAbacus.AutoABACUSModel,model_autoAbacus.AutoABACUSRunner,"Auto run abacus jobs"), 
        "06-FDMagForce":  SubParser(model_fdmagforce.FDMagForceModel,model_fdmagforce.FDMagForceModelRunner,"Do finite difference magnetic force calculation. Need mag_force_info.txt file in each example inputs."),
        "05-FDStress":  SubParser(model_fdstress.FDStressModel,model_fdstress.FDStressModelRunner,"Do finite difference stress calculation."), 
        "04-FDForce":  SubParser(model_fdforce.FDForceModel,model_fdforce.FDForceModelRunner,"Do finite difference force calculation. Need info.txt file in each example inputs."), 
        "03-Phonon": SubParser(model_phonon.PhononModel,model_phonon.PhononModelRunner,"Calculate phonon"),    
        "02-Report":SubParser(model_report.ReportModel,model_report.ReportModelRunner,"Show abacustest.html, supermetrics and metrics"),
        "01_Normal":SubParser(model_expert.ExpertModel,model_expert.ExpertModelRunner,"Run batch of abacus jobs"),
        "00-Reuse":SubParser(model_reuse.ReuseModel,model_reuse.ReuseModelRunner,"reuse model, use model to run other datasets"),
        
        #"Summary":SubParser(model_summary.SummaryModel,model_summary.SummaryModelRunner,"summary abacustest results"),

    }

if __name__ == '__main__':
    run_sp_and_exit(to_parser(), 
                    description="abacustest", 
                    version="0.1.0",
                    exception_handler=default_exception_handler)


from dp.launching.typing.basic import BaseModel, Int, String, Float, List, Optional, Union, Dict
from dp.launching.typing import (
    BaseModel,
    Set,
    Boolean,
    Field,
    )
from enum import Enum
from typing import Literal
import re,copy

class AbacusMetricEnum(String, Enum):
    AbacusMetric_metric1 = 'version'
    AbacusMetric_metric2 = 'ncore'
    AbacusMetric_metric3 = 'normal_end'
    AbacusMetric_metric4_1 = 'INPUT:ks_solver'
    AbacusMetric_metric4_2 = 'INPUT:ecutwfc'
    AbacusMetric_metric4_3 = 'INPUT:kspacing'
    AbacusMetric_metric4_4 = 'INPUT:lcao_ecut'
    AbacusMetric_metric5 = 'kpt'
    AbacusMetric_metric6_0 = 'nbase'
    AbacusMetric_metric6 = 'nbands'
    AbacusMetric_metric7 = 'converge'
    AbacusMetric_metric8 = 'total_mag'
    AbacusMetric_metric9 = 'absolute_mag'
    AbacusMetric_metric10 = 'nkstot'
    AbacusMetric_metric11 = 'ibzk'
    AbacusMetric_metric12 = 'natom'
    AbacusMetric_metric13 = 'nelec'
    AbacusMetric_metric14 = 'energy'
    AbacusMetric_metric15 = 'volume'
    AbacusMetric_metric16 = 'fft_grid'
    AbacusMetric_metric17 = 'efermi'
    AbacusMetric_metric18 = 'energy_per_atom'
    AbacusMetric_metric19 = 'stress'
    AbacusMetric_metric19_1 = 'virial'
    AbacusMetric_metric19_2 = 'pressure'
    AbacusMetric_metric20 = 'force'
    AbacusMetric_metric20_1 = 'energies'
    AbacusMetric_metric20_2 = 'forces'
    AbacusMetric_metric20_3 = 'stresses'
    AbacusMetric_metric20_4 = 'virials'
    AbacusMetric_metric20_5 = 'pressures'
    AbacusMetric_metric21 = 'band_gap'
    AbacusMetric_metric22 = 'total_time'
    AbacusMetric_metric23 = 'stress_time'
    AbacusMetric_metric24 = 'force_time'
    AbacusMetric_metric25 = 'scf_time'
    AbacusMetric_metric26 = 'scf_time_each_step'
    AbacusMetric_metric27 = 'step1_time'
    AbacusMetric_metric28 = 'scf_steps'
    AbacusMetric_metric29 = 'atom_mag'
    AbacusMetric_metric29_1 = 'atom_mag_u'
    AbacusMetric_metric29_2 = 'atom_mags'
    AbacusMetric_metric29_3 = 'atom_elec'
    AbacusMetric_metric29_4 = 'atom_orb_elec'
    AbacusMetric_metric29_5 = 'atom_elec_u'
    AbacusMetric_metric30 = 'drho'
    AbacusMetric_metric30_1 = 'drho_last'
    AbacusMetric_metric30_2 = 'denergy'
    AbacusMetric_metric30_3 = 'denergy_last'
    AbacusMetric_metric30_4 = 'denergy_womix'
    AbacusMetric_metric30_5 = 'denergy_womix_last'
    AbacusMetric_metric31 = 'lattice_constant'
    AbacusMetric_metric31_1 = 'lattice_constants'
    AbacusMetric_metric32 = 'cell'
    AbacusMetric_metric32_1 = 'cells'
    AbacusMetric_metric32_2 = 'cell_init'
    AbacusMetric_metric33 = 'coordinate'
    AbacusMetric_metric33_1 = 'coordinate_init'
    AbacusMetric_metric34 = 'element_list'
    AbacusMetric_metric35 = 'atomlabel_list'
    AbacusMetric_metric36 = 'delta_energy'
    AbacusMetric_metric37 = 'delta_energyPerAtom'
    AbacusMetric_metric38 = 'relax_converge'
    AbacusMetric_metric39 = 'relax_steps'
    AbacusMetric_metric40 = 'bda_mag_moment'
    AbacusMetric_metric41 = 'bda_bond_length'
    AbacusMetric_metric42 = 'point_group'
    AbacusMetric_metric43 = 'point_group_in_space_group'
    AbacusMetric_metric44 = 'largest_gradient'
    AbacusMetric_metric45 = 'ds_lambda_step'
    AbacusMetric_metric46 = 'ds_lambda_rms'
    AbacusMetric_metric47 = 'ds_mag'
    AbacusMetric_metric48 = 'ds_mag_force'
    
    AbacusMetric_metric_mem1 = 'mem_vkb'
    AbacusMetric_metric_mem2 = 'mem_psipw'

class SuperMetricMethodEnum(String, Enum):
    SuperMetricMethod_method1 = "iGM"
    SuperMetricMethod_method2 = "GM"
    SuperMetricMethod_method3 = "TrueRatio"
    SuperMetricMethod_method4 = "MEAN"

class SuperMetrics(BaseModel):
    #name : String
    param_name: AbacusMetricEnum 
    method: SuperMetricMethodEnum
    #normalization: Boolean

class MetricsSet(BaseModel):
    metrics: Set[AbacusMetricEnum] = Field(default=("version",'INPUT:ks_solver',"normal_end",'converge','energy','total_time','scf_steps'))
    #super_metrics: List[SuperMetrics] = Field(default=None)

class metricsSaveFileSet(BaseModel):    
    metrics_savefile: String = Field(default=None,
                                             description = "If you need to display or upload the metrics generated by executing the postdft command, please enter the file name here. \
Should be a json file, and the key is the name of the metric, and the value is the value of the metric.")
    
    super_metrics_savefile: String = Field(default=None,
                                             description = "If you need to display or upload the super metrics generated by executing the postdft command, please enter the file name here. \
Should be a json file, and the key is the name of the metric, and the value is the value of the metric.")

def convert_metrics(metrics_list):
    #the metrics in launching may have the format of KEY1:KEY2
    #this should be transfer to KEY1:{[KEY2]} in abacustest metrics
    #and transfer to KEY1/KEY2 in abacustest supermetrics
    new_metrics = []
    dict_tmp = {}
    for imetric in metrics_list:
        if ":" in imetric:
            allkeys = imetric.split(":")
            if allkeys[0] not in dict_tmp:
                dict_tmp[allkeys[0]] = []
            dict_tmp[allkeys[0]].append(allkeys[1])
        else:
            new_metrics.append(imetric)
    if dict_tmp:
        new_metrics.append(copy.deepcopy(dict_tmp))
    return new_metrics

def convert_supermetrics_metrics_name(imetric):
    if ":" in imetric:
        allkeys = imetric.split(":")
        return "%s/%s" % (allkeys[0].strip(),allkeys[1].strip())
    else:
        return imetric.strip()

def parse_metrics_set(metrics_set:MetricsSet):
    out_dict = {}
    if hasattr(metrics_set,"metrics"):
        if metrics_set.metrics == None:
            metrics = []
        else:
            metrics = list(metrics_set.metrics)
        has_super_metrics = False
        if hasattr(metrics_set,"super_metrics") and metrics_set.super_metrics != None and len(metrics_set.super_metrics) > 0:
            has_super_metrics = True
            #complete metrics
            for i in metrics_set.super_metrics:
                if i.param_name not in metrics:
                    metrics.append(i.param_name)

        #convert some special metrics (such as: KEY1:KEY2,..)
        metrics = convert_metrics(metrics)

        if len(metrics) > 0:
            #set metrics
            out_dict["metrics"] = {
                "dft_type": "abacus",
                "metrics_name": metrics,
                "save_file": "metrics_default.json",
                "modules":["bda"]
            }

            if has_super_metrics:
                out_dict["super_metrics"] = [{
                    "save_file": "superMetrics.json",
                    "result_file": ["metrics_default.json"],
                    "metrics":[],
                    "outparams":[]
                }]  
                for i in metrics_set.super_metrics:
                    metric_name = convert_supermetrics_metrics_name(i.param_name)
                    out_dict["super_metrics"][-1]["metrics"].append({
                        "name": f"{i.method}({metric_name})" ,
                        "param_name": metric_name,
                        "method": i.method,
                        "normalization": False})
                    out_dict["super_metrics"][-1]["outparams"].append([metric_name, [metric_name], -1])
                    
    if hasattr(metrics_set,"metrics_savefile"):
        if metrics_set.metrics_savefile != None and metrics_set.metrics_savefile.strip() != "":
            if "metrics" not in out_dict:
                out_dict["metrics"] = {}
            out_dict["metrics"]["value_from_file"] = metrics_set.metrics_savefile.strip()
        if metrics_set.super_metrics_savefile != None and metrics_set.super_metrics_savefile.strip() != "":
            if "super_metrics" not in out_dict:
                out_dict["super_metrics"] = [{}]
            out_dict["super_metrics"][-1]["value_from_file"] = metrics_set.super_metrics_savefile.strip()
    
    return out_dict

